(* \comment{ *)

exception Found
exception BunchErrorException
exception BunchMatchException

(* end commented fragment } *)
(*\codefrag\typenode*)
type 'a node = 
    Leaf of 'a | Add | Mul
  | End | OBracket | CBracket
(*\codefrag\datatype*)
type 'a bunch = 
    Empty
  | L of 'a
  | MB of 'a bunch list 
  | AB of 'a bunch list 
(*\codefrag\mjoin*)
let mjoin ba bb =
  let f a b = 
    match (a,b) with
        (MB m , MB n) -> m@n
      | (MB m, _ ) -> m@[b]
      | ( _ , MB m) -> a::m
      | _  -> a::[b]
  in
      if ba = Empty then bb
      else if bb = Empty then ba
      else MB (f ba bb)
(*\codefrag\ajoin*)
let ajoin ba bb =
  let f a b = 
    match (a,b) with
        (AB m , AB n) -> m@n
      | (AB m, _ ) -> m@[b]
      | ( _ , AB m) -> a::m
      | _  -> a::[b]
  in
      if ba = Empty then bb
      else if bb = Empty then ba
      else AB (f ba bb)
(*\codefrag\arity*)
let rec arity b =
  let f l = (List.length l)::List.concat (List.map arity l)
  in
      match b with 
          Empty -> [0]
        | L a -> [1]
        | MB m -> 1::(f m)
        | AB m -> f m
(*\codefrag\proc*)
let rec proc f = 
  let g j c l = j (proc f (List.hd l))(proc f (c (List.tl l)))
  and fmb l = MB l
  and fab l = AB l
  in function
      Empty -> Empty
    | L a -> f a
    | MB (h::[]) -> proc f h
    | MB m -> g mjoin fmb m
    | AB (h::[]) -> proc f h
    | AB m -> g ajoin fab m
(*\codefrag\bmap*)
let rec bmap f =
  let g l = List.map (bmap f) l
  in function
      Empty -> Empty
    | L a -> L (f a)
    | MB m -> MB (g m)
    | AB m -> AB (g m)
(*\codefrag\bzip*)
let rec bzip b1 b2 =
  match (b1,b2) with
      (Empty,Empty) -> Empty
    | (L a , L b) -> L (a,b)
    | (MB (m1::[]), MB (m2::[])) -> bzip m1 m2
    | (MB (m1::tl1), MB (m2::tl2)) -> 
          mjoin (bzip m1 m2) (bzip (MB tl1) (MB tl2))
    | (AB (m1::[]), AB (m2::[])) -> bzip m1 m2
    | (AB (m1::tl1), AB (m2::tl2)) -> 
          ajoin (bzip m1 m2) (bzip (AB tl1) (AB tl2))
    | _ -> raise (Failure "Mismatch in bzip")
(*\codefrag\biter*)
let rec biter f =
  let g l =  List.iter (biter f) l
  in function
      Empty -> ()
    | L a -> f a
    | MB m -> g m
    | AB m -> g m
(*\codefrag\bexists*)
let bexists f bunch =
  let g a = if f a then raise Found else ()
  and l = ref false in
      begin try biter g bunch;()
      with Found -> l := true
      end;
      !l
(*\codefrag\mem*)
let mem a bunch =
  let f b = if a == b then raise Found else ()
  and l = ref false in
      begin try biter f bunch;()
      with Found -> l := true
      end;
      !l
(*\codefrag\subbunch*)
let subbunch a bunch =
  let f b = if not (mem b bunch) then raise Not_found else () 
  and l = ref true in
      begin try biter f a;()
      with Not_found -> l := false
      end;
      !l
(*\codefrag\intersection*)
let rec intersection a = function
    Empty -> Empty
  | (L c) as bunch -> if mem c a then bunch else Empty
  | MB (h::[]) -> intersection h a
  | MB (h::tl) -> mjoin (intersection h a) (intersection (MB tl) a)
  | AB (h::[]) -> intersection h a
  | AB (h::tl) -> ajoin (intersection h a) (intersection (AB tl) a)
  | _ -> raise BunchErrorException
(*\codefrag\check*)
let check sb bunch =
  let f b = if mem b bunch then raise Found else () 
  and l = ref true in
      begin try biter f sb;()
      with Found -> l := false
      end; !l
(*\codefrag\remfirst*)
let rec remfirst a bunch = 
  let f b = if b == a then Empty else L b in
      proc f bunch
(*\codefrag\bunchaxiom*)
let rec axiom a = function
    Empty -> Empty
  | (L c) as bunch -> if c == a then Empty else bunch
  | MB (h::[]) ->
        axiom a h
  | MB (h::tl) ->
        let ntl = 
          if List.length tl = 1 then List.hd tl
          else MB tl
        in
            if mem a h then 
                mjoin (axiom a h) ntl
            else mjoin h (axiom a ntl)
  | AB (h::[]) ->
        axiom a h
  | (AB (h::tl)) as bunch ->
        if not (mem a bunch) then bunch else
        if mem a h then axiom a h
        else axiom a (AB tl)
  | _ -> raise BunchErrorException
(*\codefrag\replace*)
let rec replace clause quantified = function
      Empty -> raise (Failure "empty in replace")
    | (L a) as c -> if a == quantified then clause else c
    | MB (h::[]) -> replace clause quantified h
    | MB (h::tl) -> 
          if mem quantified h then MB (replace clause quantified h::tl)
          else mjoin h (replace clause quantified (MB tl))
    | AB (h::[]) -> replace clause quantified h
    | AB (h::tl) -> 
          if mem quantified h then AB (replace clause quantified h::tl)
          else ajoin h (replace clause quantified (AB tl))
    | _ -> raise BunchErrorException
(*\codefrag\subtract*)
let rec subtract sub bunch = 
  match sub with
      Empty -> bunch
    | L a -> axiom a bunch
    | MB (h::[]) -> 
          subtract h bunch
    | MB (h::tl) -> 
          subtract h (subtract (MB tl) bunch)
    | AB (h::tl) -> subtract h bunch
    | _ -> raise BunchErrorException
(*\codefrag\getabunch*)
let rec getabunch a = function
    Empty -> raise (Failure "empty in getabunch")
  | (L c) as bunch -> if c == a then bunch
            else raise (Failure "-")
  | MB (h::[]) ->
        getabunch a h
  | MB (h::tl) ->
        if mem a h then getabunch a h
        else getabunch a (MB tl)
  | (AB list) as bunch ->
        let rec f = function 
            L c -> c == a
          | _ -> false
        and g  = function
            (h::[]) -> getabunch a h
          | (h::tl) ->
                if mem a h then getabunch a h
                else g tl
          | _ -> raise BunchErrorException
        in 
            if (List.fold_right (or) (List.map f list) false) 
            then bunch
            else g list
  | _ -> raise BunchErrorException
(*\codefrag\mnw*)
let rec mnw a = function
   |  MB (h::[]) -> if mem a h then mnw a h else h
   |  MB (h::tl) -> if mem a h then mjoin (mnw a h) (MB tl)
                    else mjoin h (mnw a (MB tl))
   |  AB (h::[]) -> mnw a h
   |  AB (h::tl) -> if mem a h then mnw a h
                    else mnw a (AB tl)
   | cl -> cl
(*\codefrag\buildbunch*)
let rec build_bunch = parser
    [< e = build_mult; e1 = build_rest_add e >] -> e1
and build_rest_add e = parser
    [< 'Add; e1 = build_mult; 
        e2 = build_rest_add (ajoin e e1) >] -> e2
  | [< >] -> e
and build_mult = parser
    [< e = build_leaf; e1 = build_rest_mul e >] -> e1
and build_rest_mul e = parser
    [< 'Mul; e1 = build_leaf;
        e2 = build_rest_mul (mjoin e e1) >] -> e2
  | [< >] -> e
and build_leaf = parser
    [< 'Leaf a >] -> L a
  | [< 'OBracket; e = build_bunch; 'CBracket >] -> e
  | [< 'End >] -> Empty
(* \comment{

for printing; f: 'a -> string *)
let rec bunch_to_string f = 
  let rec g c = function 
      [] -> ""
    | (h::[]) -> bunch_to_string f h
    | (h::tl) -> bunch_to_string f h ^ c ^ g c tl
  in function
    Empty -> "Empty"
  | L a -> f a
  | MB l -> "("^(g "," l) ^ ")"
  | AB l -> "("^(g ";" l)^")"

let print_bunch f b = print_string(bunch_to_string f b)

let pib = print_bunch string_of_int
let psb = print_bunch (function x -> x)

let lexer = Genlex.make_lexer["(";")";",";";";":";"."]

let rec parse s = match s with parser
    [< 'Genlex.Kwd "(" >] -> [< 'OBracket ; parse s >]
  | [< 'Genlex.Kwd ")" >] -> [< 'CBracket ; parse s >]
  | [< 'Genlex.Kwd "," >] -> [< 'Mul ; parse s >]
  | [< 'Genlex.Kwd ";" >] -> [< 'Add ; parse s >]
  | [< 'Genlex.Ident i >] -> [< 'Leaf i ; parse s >]
  | [< 'Genlex.Int i >] -> [< 'Leaf (string_of_int i) ; parse s >]
  | [< >] -> [< 'End >]

let make s = build_bunch (parse (lexer (Stream.of_string s)))


let a = make "a"
let b = make "b"
let c = make "c"
let d = make "d"
let e = make "e"
let f = make "f"
let ab = ajoin a b
let cd = ajoin c d
let ef = ajoin e f
let abc = mjoin ab c
let acd = mjoin a cd
let abcd = mjoin ab cd
let cdef = mjoin cd ef

let p a b c = 
   let ps = print_string in
   ps "("; psb a; ps " , "; psb b; ps ") - ";psb c; ps " = ";
   psb (subtract c (mjoin a b));
   print_string "\n";
   psb a; ps "' , ("; psb b; ps " - ";psb c; ps ") = ";
   psb (mjoin (mjoin (subtract c a) c) (subtract c b));
   print_string "\n"

(*finalcomment*)

(* end of skipped code } *)

